import logging

from jeeves.core.tasks import PROVIDED_TASKS


class Singleton(type):
    def __init__(cls, name, bases, attrs, **kwargs):
        super().__init__(name, bases, attrs, **kwargs)
        cls._instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__call__(*args, **kwargs)
        return cls._instance


class TaskRegistry(metaclass=Singleton):
    tasks = {}

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    @classmethod
    def autodiscover(cls):
        """Loads all provided tasks."""
        # TODO: Third party plugins
        registry = cls()
        for task_namespace in PROVIDED_TASKS:
            registry.register_task(cls.get_task_cls(task_namespace))

    @classmethod
    def register_task(cls, task_cls):
        registry = cls()
        # namespace = task_cls.id
        namespace = f"{task_cls.__module__}:{task_cls.__name__}"

        if namespace in registry.tasks:
            raise cls.TaskNamespaceConflict(
                f"Namespace {namespace} is already registered"
            )

        registry.tasks[namespace] = task_cls

    @classmethod
    def get_task_cls(cls, namespace, **kwargs):
        # Right now tasks are being imported and returned dinamically because it's easier,
        # but we will need a way of autodiscover all tasks (or register them manually) and
        # referencing them on a list so the user knows which tasks are available.

        modulename, clsname = namespace.split(":")
        try:
            module = __import__(f"{modulename}", fromlist=(clsname,), level=0)
            task_cls = getattr(module, clsname)
            return task_cls
        except ModuleNotFoundError as error:
            raise cls.TaskDoesNotExist(f"Error importing task {namespace}: {error.msg}")

    class TaskNamespaceConflict(Exception):
        pass

    class TaskDoesNotExist(Exception):
        """
        Used when there's a problem retrieving a task. More info will be available on the message.
        """

        pass
